"""OpenAI client."""
import logging
import os
from typing import Any, Dict, Optional

from manifest.clients.client import Client
from manifest.request import LMRequest

logger = logging.getLogger(__name__)

OPENAI_ENGINES = {
    "text-davinci-003",
    "text-davinci-002",
    "text-davinci-001",
    "davinci",
    "curie",
    "ada",
    "babbage",
    "text-curie-001",
    "text-babbage-001",
    "text-ada-001",
    "code-davinci-002",
    "code-cushman-001",
}


class OpenAIClient(Client):
    """OpenAI client."""

    # User param -> (client param, default value)
    PARAMS = {
        "engine": ("model", "text-davinci-003"),
        "temperature": ("temperature", 1.0),
        "max_tokens": ("max_tokens", 10),
        "n": ("n", 1),
        "top_p": ("top_p", 1.0),
        "top_k": ("best_of", 1),
        "stop_sequences": ("stop", None),  # OpenAI doesn't like empty lists
        "presence_penalty": ("presence_penalty", 0.0),
        "frequency_penalty": ("frequency_penalty", 0.0),
        "client_timeout": ("client_timeout", 60),  # seconds
    }
    REQUEST_CLS = LMRequest

    def connect(
        self,
        connection_str: Optional[str] = None,
        client_args: Dict[str, Any] = {},
    ) -> None:
        """
        Connect to the OpenAI server.

        connection_str is passed as default OPENAI_API_KEY if variable not set.

        Args:
            connection_str: connection string.
            client_args: client arguments.
        """
        self.api_key = os.environ.get("OPENAI_API_KEY", connection_str)
        if self.api_key is None:
            raise ValueError(
                "OpenAI API key not set. Set OPENAI_API_KEY environment "
                "variable or pass through `client_connection`."
            )
        self.host = "https://api.openai.com/v1"
        for key in self.PARAMS:
            setattr(self, key, client_args.pop(key, self.PARAMS[key][1]))
        if getattr(self, "engine") not in OPENAI_ENGINES:
            raise ValueError(
                f"Invalid engine {getattr(self, 'engine')}. Must be {OPENAI_ENGINES}."
            )

    def close(self) -> None:
        """Close the client."""
        pass

    def get_generation_url(self) -> str:
        """Get generation URL."""
        return self.host + "/completions"

    def get_generation_header(self) -> Dict[str, str]:
        """
        Get generation header.

        Returns:
            header.
        """
        return {"Authorization": f"Bearer {self.api_key}"}

    def supports_batch_inference(self) -> bool:
        """Return whether the client supports batch inference."""
        return True

    def get_model_params(self) -> Dict:
        """
        Get model params.

        By getting model params from the server, we can add to request
        and make sure cache keys are unique to model.

        Returns:
            model params.
        """
        return {"model_name": "openai", "engine": getattr(self, "engine")}



"""Dummy client."""
import logging
from typing import Any, Callable, Dict, Optional, Tuple

from manifest.clients.client import Client
from manifest.request import LMRequest, Request

logger = logging.getLogger(__name__)


class DummyClient(Client):
    """Dummy client."""

    # User param -> (client param, default value)
    PARAMS = {
        "n": ("num_results", 1),
    }
    REQUEST_CLS = LMRequest

    def connect(
        self,
        connection_str: Optional[str] = None,
        client_args: Dict[str, Any] = {},
    ) -> None:
        """
        Connect to dummpy server.

        This is a dummy client that returns identity responses. Used for testing.

        Args:
            connection_str: connection string.
            client_args: client arguments.
        """
        for key in self.PARAMS:
            setattr(self, key, client_args.pop(key, self.PARAMS[key][1]))

    def close(self) -> None:
        """Close the client."""
        pass

    def get_generation_url(self) -> str:
        """Get generation URL."""
        return "dummy"

    def supports_batch_inference(self) -> bool:
        """Return whether the client supports batch inference."""
        return True

    def get_generation_header(self) -> Dict[str, str]:
        """
        Get generation header.

        Returns:
            header.
        """
        return {}

    def get_model_params(self) -> Dict:
        """
        Get model params.

        By getting model params from the server, we can add to request
        and make sure cache keys are unique to model.

        Returns:
            model params.
        """
        return {"engine": "dummy"}

    def get_request(self, request: Request) -> Tuple[Callable[[], Dict], Dict]:
        """
        Get request string function.

        Args:
            request: request.

        Returns:
            request function that takes no input.
            request parameters as dict.
        """
        if isinstance(request.prompt, list):
            num_results = len(request.prompt)
        else:
            num_results = 1
        request_params = request.to_dict(self.PARAMS)

        def _run_completion() -> Dict:
            return {
                "choices": [{"text": "hello"}]
                * int(request_params["num_results"])
                * num_results
            }

        return _run_completion, request_params

    def get_score_prompt_request(
        self,
        request: Request,
    ) -> Tuple[Callable[[], Dict], Dict]:
        """
        Get the logit score of the prompt via a forward pass of the model.

        Args:
            request: request.

        Returns:
            request function that takes no input.
            request parameters as dict.
        """
        if isinstance(request.prompt, list):
            num_results = len(request.prompt)
        else:
            num_results = 1
        request_params = {"prompt": request.prompt}

        def _run_completion() -> Dict:
            return {
                "choices": [
                    {
                        "text": request.prompt
                        if isinstance(request.prompt, str)
                        else request.prompt[i],
                        "logprob": 0.3,
                    }
                    for i in range(num_results)
                ]
            }

        return _run_completion, request_params



"""Manifest init."""
from manifest.manifest import Manifest
from manifest.request import Request
from manifest.response import Response
from manifest.session import Session

__all__ = ["Manifest", "Response", "Session"]



"""Manifest class."""
import logging
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np

from manifest.caches.noop import NoopCache
from manifest.caches.redis import RedisCache
from manifest.caches.sqlite import SQLiteCache
from manifest.clients.ai21 import AI21Client
from manifest.clients.chatgpt import ChatGPTClient
from manifest.clients.cohere import CohereClient
from manifest.clients.diffuser import DiffuserClient
from manifest.clients.dummy import DummyClient
from manifest.clients.huggingface import HuggingFaceClient
from manifest.clients.openai import OpenAIClient
from manifest.clients.toma import TOMAClient
from manifest.clients.toma_diffuser import TOMADiffuserClient
from manifest.request import Request
from manifest.response import Response
from manifest.session import Session

logging.getLogger("openai").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)

CLIENT_CONSTRUCTORS = {
    "openai": OpenAIClient,
    "chatgpt": ChatGPTClient,
    "cohere": CohereClient,
    "ai21": AI21Client,
    "huggingface": HuggingFaceClient,
    "diffuser": DiffuserClient,
    "dummy": DummyClient,
    "toma": TOMAClient,
    "tomadiffuser": TOMADiffuserClient,
}

CACHE_CONSTRUCTORS = {
    "redis": RedisCache,
    "sqlite": SQLiteCache,
    "noop": NoopCache,
}


class Manifest:
    """Manifest session object."""

    def __init__(
        self,
        client_name: str = "openai",
        client_connection: Optional[str] = None,
        cache_name: str = "noop",
        cache_connection: Optional[str] = None,
        session_id: Optional[str] = None,
        stop_token: str = "",
        **kwargs: Any,
    ):
        """
        Initialize manifest.

        Args:
            client_name: name of client.
            client_connection: connection string for client.
            cache_name: name of cache.
            cache_connection: connection string for cache.
            session_id: session id for user session cache.
                        None (default) means no session logging.
                        "_default" means generate new session id.
            stop_token: stop token prompt generation.
                        Can be overridden in run

        Remaining kwargs sent to client and cache.
        """
        if client_name not in CLIENT_CONSTRUCTORS:
            raise ValueError(
                f"Unknown client name: {client_name}. "
                f"Choices are {list(CLIENT_CONSTRUCTORS.keys())}"
            )
        if cache_name not in CACHE_CONSTRUCTORS:
            raise ValueError(
                f"Unknown cache name: {cache_name}. "
                f"Choices are {list(CACHE_CONSTRUCTORS.keys())}"
            )
        self.client_name = client_name
        # Must pass kwargs as dict for client "pop" methods removed used arguments
        self.cache = CACHE_CONSTRUCTORS[cache_name](  # type: ignore
            cache_connection, self.client_name, cache_args=kwargs
        )
        self.client = CLIENT_CONSTRUCTORS[self.client_name](  # type: ignore
            client_connection, client_args=kwargs
        )
        if session_id is not None:
            if self.client_name == "diffuser":
                raise NotImplementedError(
                    "Session logging not implemented for Diffuser client."
                )
            if session_id == "_default":
                # Set session_id to None for Session random id
                session_id = None
            self.session = Session(session_id)
        else:
            self.session = None
        if len(kwargs) > 0:
            raise ValueError(f"{list(kwargs.items())} arguments are not recognized.")

        self.stop_token = stop_token

    def close(self) -> None:
        """Close the client and cache."""
        self.client.close()
        self.cache.close()

    def change_client(
        self,
        client_name: Optional[str] = None,
        client_connection: Optional[str] = None,
        stop_token: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Change manifest client.

        Args:
            client_name: name of client.
            client_connection: connection string for client.
            stop_token: stop token prompt generation.
                        Can be overridden in run

        Remaining kwargs sent to client.
        """
        if client_name:
            if client_name not in CLIENT_CONSTRUCTORS:
                raise ValueError(
                    f"Unknown client name: {client_name}. "
                    f"Choices are {list(CLIENT_CONSTRUCTORS.keys())}"
                )
            self.client_name = client_name
            self.client = CLIENT_CONSTRUCTORS[client_name](  # type: ignore
                client_connection, client_args=kwargs
            )
            if len(kwargs) > 0:
                raise ValueError(
                    f"{list(kwargs.items())} arguments are not recognized."
                )

        if stop_token is not None:
            self.stop_token = stop_token

    def _validate_kwargs(self, kwargs: Dict, request_params: Request) -> None:
        """Validate kwargs.

        Args:
            kwargs: kwargs to validate.
            request_params: request object to validate against.
        """
        # Check for invalid kwargs
        non_request_kwargs = [
            (k, v) for k, v in kwargs.items() if k not in request_params.__dict__
        ]
        if len(non_request_kwargs) > 0:
            raise ValueError(
                f"{list(non_request_kwargs)} arguments are not recognized."
            )

        # Warn for valid but unused kwargs
        request_unused_kwargs = [
            (k, v) for k, v in kwargs.items() if k not in non_request_kwargs
        ]
        if len(request_unused_kwargs) > 0:
            logger.warning(f"{list(request_unused_kwargs)} arguments are unused.")
        return

    def run(
        self,
        prompt: Union[str, List[str]],
        overwrite_cache: bool = False,
        run_id: Optional[str] = None,
        stop_token: Optional[str] = None,
        return_response: bool = False,
        **kwargs: Any,
    ) -> Union[str, List[str], np.ndarray, List[np.ndarray], Response]:
        """
        Run the prompt.

        Args:
            prompt: prompt(s) to run.
            overwrite_cache: whether to overwrite cache.
            run_id: run id for cache to repeat same run.
            stop_token: stop token for prompt generation.
                        Default is self.stop_token.
                        "" for no stop token.
            return_response: whether to return Response object.

        Returns:
            response from prompt.
        """
        is_batch = isinstance(prompt, list)

        stop_token = stop_token if stop_token is not None else self.stop_token
        # Must pass kwargs as dict for client "pop" methods removed used arguments
        request_params = self.client.get_request_params(prompt, kwargs)
        # Avoid nested list of results - enforce n = 1 for batch
        if is_batch and request_params.n > 1:
            raise ValueError("Batch mode does not support n > 1.")
        possible_request, full_kwargs = self.client.get_request(request_params)

        self._validate_kwargs(kwargs, request_params)
        # Create cacke key
        cache_key = full_kwargs.copy()
        # Make query model dependent
        cache_key.update(self.client.get_model_params())
        if run_id:
            cache_key["run_id"] = run_id
        response_obj = self.cache.get(cache_key, overwrite_cache, possible_request)
        # Log session dictionary values
        if self.session:
            self.session.log_query(cache_key, response_obj.to_dict())
        # Extract text results
        if return_response:
            return response_obj
        else:
            return response_obj.get_response(stop_token, is_batch)

    def score_prompt(
        self,
        prompt: Union[str, List[str]],
        overwrite_cache: bool = False,
        **kwargs: Any,
    ) -> Dict:
        """
        Score the prompt via forward pass of the model - no sampling or generation.

        Returns the response object with logits of the prompt.

        Prompt scoring is not part of a session cache.

        Args:
            prompt: prompt(s) to run.
            overwrite_cache: whether to overwrite cache.

        Returns:
            response from prompt.
        """
        # Must pass kwargs as dict for client "pop" methods removed used arguments
        request_params = self.client.get_request_params(prompt, kwargs)

        if request_params.n > 1:
            raise ValueError("Sequence scoring does not support n > 1.")

        try:
            possible_request, full_kwargs = cast(
                HuggingFaceClient, self.client
            ).get_score_prompt_request(request_params)
        except AttributeError:
            raise ValueError("`score_prompt` only supported for HF models.")

        self._validate_kwargs(kwargs, request_params)
        # Create cacke key
        cache_key = full_kwargs.copy()
        # Make query model dependent
        cache_key.update(self.client.get_model_params())
        response_obj = self.cache.get(cache_key, overwrite_cache, possible_request)
        return response_obj.to_dict()

    def get_last_queries(
        self,
        last_n: int = -1,
        return_raw_values: bool = False,
        stop_token: Optional[str] = None,
    ) -> List[Tuple[Any, Any]]:
        """
        Get last n queries from current session.

        If last_n is -1, return all queries. By default will only return the
        prompt text and result text unles return_raw_values is False.

        Args:
            last_n: last n queries.
            return_raw_values: whether to return raw values as dicts.
            stop_token: stop token for prompt results to be applied to all results.

        Returns:
            last n list of queries and outputs.
        """
        if self.session is None:
            raise ValueError(
                "Session was not initialized. Set `session_id` when loading Manifest."
            )
        stop_token = stop_token if stop_token is not None else self.stop_token
        last_queries = self.session.get_last_queries(last_n)
        if not return_raw_values:
            last_queries = [
                (
                    query["prompt"],
                    Response.from_dict(response).get_response(
                        stop_token, is_batch=isinstance(query["prompt"], list)
                    ),
                )  # type: ignore
                for query, response in last_queries
            ]
        return last_queries

    def open_explorer(self) -> None:
        """Open the explorer for jupyter widget."""
        # Open explorer
        # TODO: implement
        pass



"""User query session logging."""
import logging
import os
import sqlite3
import uuid
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from manifest.caches.serializers import Serializer

logging.getLogger("sqlitedict").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


class Session:
    """A user session for caching requests."""

    def __init__(self, session_id: Optional[str] = None) -> None:
        """
        Initialize session.

        If session_id already exists, will append to existing session.

        Args:
            session_id: session id.

        """
        manifest_home = Path(os.environ.get("MANIFEST_HOME", Path.home()))
        self.db_file = manifest_home / ".manifest" / "session.db"
        self.db_file.parent.mkdir(parents=True, exist_ok=True)
        self.conn = sqlite3.connect(str(self.db_file))
        self.serializer = Serializer()
        self._create_table()
        if not session_id:
            self.session_id = str(uuid.uuid4())
            self.query_id = 0
        else:
            self.session_id = session_id
            self.query_id = self._get_latest_query_id(self.session_id)
            self.query_id += 1
        logger.info(f"Starting session {self.session_id}")
        return

    def close(self) -> None:
        """Close the client."""
        self.conn.close()

    @classmethod
    def get_session_keys(cls, db_file: Path) -> List[str]:
        """Get available session keys from cached file."""
        try:
            conn = sqlite3.connect(str(db_file))
            query = """SELECT DISTINCT session_id FROM queries"""
            cur = conn.cursor()
            res = cur.execute(query)
            return [x[0] for x in res.fetchall()]
        except sqlite3.OperationalError:
            logger.info(
                "There is no database with the 'queries' table. "
                "Are you sure you are using the right session file"
            )
            return []

    def _execute_query(self, query: str, *args: Any) -> Any:
        """
        Execute query with optional args.

        Args:
            query: query to execute.
        """
        cur = self.conn.cursor()
        res = cur.execute(query, args)
        self.conn.commit()
        return res

    def _create_table(self) -> None:
        """Create table if not exists."""
        query = """CREATE TABLE IF NOT EXISTS queries (
            query_id integer NOT NULL,
            session_id text NOT NULL,
            query_key text NOT NULL,
            response_key text NOT NULL
        );"""
        self._execute_query(query)
        return

    def _get_latest_query_id(self, session_id: str) -> int:
        """
        Get latest query id issued if resuming session.

        If no session_id, return -1.

        Args:
            session_id: session id.

        Returns:
            latest query id.
        """
        query = """SELECT query_id
                    FROM queries
                    WHERE session_id = ?
                    ORDER BY query_id DESC LIMIT 1"""
        res = self._execute_query(query, session_id).fetchone()
        if res:
            return res[0]
        return -1

    def log_query(
        self, query_key: Dict[str, Any], response_key: Dict[str, Any]
    ) -> None:
        """
        Log the query and response.

        Args:
            query_key: query of user (dump of request params).
            response_key: response of server (dump of response).
        """
        query = """INSERT INTO queries VALUES (?, ?, ?, ?);"""
        self._execute_query(
            query,
            self.query_id,
            self.session_id,
            self.serializer.request_to_key(query_key),
            self.serializer.response_to_key(response_key),
        )
        self.query_id += 1
        return

    def get_last_queries(
        self, last_n: int = -1
    ) -> List[Tuple[Dict[str, Any], Dict[str, Any]]]:
        """
        Get last n queries from current session.

        If last_n is -1, return all queries.

        Args:
            last_n: last n queries.

        Returns:
            last n list of queries and outputs.
        """
        first_query = self.query_id - last_n if last_n > 0 else -1
        query = """SELECT query_key, response_key
                    FROM queries
                    WHERE session_id = ? AND query_id >= ?
                    ORDER BY query_id;"""
        res = self._execute_query(query, self.session_id, first_query)
        parsed_res = [
            (
                self.serializer.key_to_request(pair[0]),
                self.serializer.key_to_response(pair[1]),
            )
            for pair in res.fetchall()
        ]
        return parsed_res



"""Client class."""
import logging
import os
from typing import Any, Callable, Dict, Optional, Tuple

from pyChatGPT import ChatGPT

from manifest.clients.client import Client
from manifest.request import LMRequest, Request

logger = logging.getLogger(__name__)


class ChatGPTClient(Client):
    """ChatGPT Client class."""

    # No params for ChatGPT
    PARAMS: Dict[str, Tuple[str, Any]] = {}
    REQUEST_CLS = LMRequest

    def connect(
        self, connection_str: Optional[str], client_args: Dict[str, Any]
    ) -> None:
        """
        Connect to ChatGPT.

        We use https://github.com/terry3041/pyChatGPT.

        Arsg:
            connection_str: connection string.
            client_args: client arguments.
        """
        self.session_key = os.environ.get("CHATGPT_SESSION_KEY", connection_str)
        if self.session_key is None:
            raise ValueError(
                "ChatGPT session key not set. Set CHATGPT_SESSION_KEY environment "
                "variable or pass through `client_connection`. "
                "For details, see https://github.com/terry3041/pyChatGPT "
                "and go through instructions for getting a session key."
            )
        self.host = None
        for key in self.PARAMS:
            setattr(self, key, client_args.pop(key, self.PARAMS[key][1]))
        self._chat_session = ChatGPT(self.session_key, verbose=False)

    def close(self) -> None:
        """Close the client."""
        self._chat_session = None

    def clear_conversations(self) -> None:
        """Clear conversations.

        Only works for ChatGPT.
        """
        self._chat_session.clear_conversations()

    def get_generation_url(self) -> str:
        """Get generation URL."""
        return ""

    def get_generation_header(self) -> Dict[str, str]:
        """
        Get generation header.

        Returns:
            header.
        """
        return {}

    def supports_batch_inference(self) -> bool:
        """Return whether the client supports batch inference."""
        return False

    def get_model_params(self) -> Dict:
        """
        Get model params.

        By getting model params from the server, we can add to request
        and make sure cache keys are unique to model.

        Returns:
            model params.
        """
        return {"model_name": "chatgpt", "engine": "chatgpt"}

    def format_response(self, response: Dict) -> Dict[str, Any]:
        """
        Format response to dict.

        Args:
            response: response

        Return:
            response as dict
        """
        return {
            "model": "chatgpt",
            "choices": [
                {
                    "text": response["message"],
                }
            ],
        }

    def get_request(self, request: Request) -> Tuple[Callable[[], Dict], Dict]:
        """
        Get request string function.

        Args:
            request: request.

        Returns:
            request function that takes no input.
            request parameters as dict.
        """
        if isinstance(request.prompt, list):
            raise ValueError("ChatGPT does not support batch inference.")

        prompt = str(request.prompt)
        request_params = request.to_dict(self.PARAMS)

        def _run_completion() -> Dict:
            try:
                res = self._chat_session.send_message(prompt)
            except Exception as e:
                logger.error(f"ChatGPT error {e}.")
                raise e
            return self.format_response(res)

        return _run_completion, request_params



"""Client response."""
import json
from typing import Any, Dict, List, Union

import numpy as np


class NumpyArrayEncoder(json.JSONEncoder):
    """Numpy array encoder."""

    def default(self, obj: Any) -> str:
        """Encode numpy array."""
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class Response:
    """Response class."""

    def __init__(
        self,
        response: Dict,
        cached: bool,
        request_params: Dict,
        generation_key: str = "choices",
        logits_key: str = "logprobs",
        item_key: str = "text",
    ):
        """
        Initialize response.

        Args:
            response: response dict.
            cached: whether response is cached.
            request_params: request parameters.
            generation_key: key for generation results.
            logits_key: key for logits.
            item_key: key for item in the generations.
        """
        self.generation_key = generation_key
        self.logits_key = logits_key
        self.item_key = item_key
        self.item_dtype = None
        if isinstance(response, dict):
            self._response = response
        else:
            raise ValueError(f"Response must be dict. Response is\n{response}.")
        if (
            (self.generation_key not in self._response)
            or (not isinstance(self._response[self.generation_key], list))
            or (len(self._response[self.generation_key]) <= 0)
        ):
            raise ValueError(
                "Response must be serialized to a dict with a nonempty"
                f" list of choices. Response is\n{self._response}."
            )
        if self.item_key not in self._response[self.generation_key][0]:
            raise ValueError(
                "Response must be serialized to a dict with a "
                f"list of choices with {self.item_key} field"
            )
        if (
            self.logits_key in self._response[self.generation_key][0]
            and self._response[self.generation_key][0][self.logits_key]
        ):
            if not isinstance(
                self._response[self.generation_key][0][self.logits_key], list
            ):
                raise ValueError(
                    "Response must be serialized to a dict with a "
                    "list of choices with logprobs field"
                )
        if isinstance(
            self._response[self.generation_key][0][self.item_key], np.ndarray
        ):
            self.item_dtype = str(
                self._response[self.generation_key][0][self.item_key].dtype
            )
        self._cached = cached
        self._request_params = request_params

    def is_cached(self) -> bool:
        """Check if response is cached."""
        return self._cached

    def get_request(self) -> Dict:
        """Get request parameters."""
        return self._request_params

    def get_json_response(self) -> Dict:
        """Get response dict without parsing."""
        return self._response

    def get_response(
        self, stop_token: str = "", is_batch: bool = False
    ) -> Union[str, List[str], np.ndarray, List[np.ndarray]]:
        """
        Get all results from response.

        Args:
            stop_token: stop token for string generation
            is_batch: whether response is batched
        """
        process_result = (
            lambda x: x.strip().split(stop_token)[0] if stop_token else x.strip()
        )
        extracted_items = [
            choice[self.item_key] for choice in self._response[self.generation_key]
        ]
        if len(extracted_items) == 0:
            return None
        if isinstance(extracted_items[0], str):
            processed_results = list(map(process_result, extracted_items))
        else:
            processed_results = extracted_items
        if len(processed_results) == 1 and not is_batch:
            return processed_results[0]
        else:
            return processed_results

    def serialize(self) -> str:
        """
        Serialize response to string.

        Returns:
            serialized response.
        """
        return json.dumps(self.to_dict(), sort_keys=True, cls=NumpyArrayEncoder)

    @classmethod
    def deserialize(cls, value: str) -> "Response":
        """
        Deserialize string to response.

        Args:
            value: serialized response.

        Returns:
            serialized response.
        """
        deserialized = json.loads(value)
        item_dtype = deserialized["item_dtype"]
        if item_dtype:
            for choice in deserialized["response"][deserialized["generation_key"]]:
                choice[deserialized["item_key"]] = np.array(
                    choice[deserialized["item_key"]]
                ).astype(item_dtype)
        return cls(
            deserialized["response"],
            deserialized["cached"],
            deserialized["request_params"],
            generation_key=deserialized["generation_key"],
            logits_key=deserialized["logits_key"],
            item_key=deserialized["item_key"],
        )

    def to_dict(self) -> Dict:
        """
        Get dictionary representation of response.

        Returns:
            dictionary representation of response.
        """
        return {
            "generation_key": self.generation_key,
            "logits_key": self.logits_key,
            "item_key": self.item_key,
            "item_dtype": self.item_dtype,
            "response": self._response,
            "cached": self._cached,
            "request_params": self._request_params,
        }

    @classmethod
    def from_dict(cls, response: Dict) -> "Response":
        """
        Create response from dictionary.

        Args:
            response: dictionary representation of response.

        Returns:
            response.
        """
        return cls(
            response["response"],
            response["cached"],
            response["request_params"],
            generation_key=response["generation_key"],
            logits_key=response["logits_key"],
            item_key=response["item_key"],
        )

    def __str__(self) -> str:
        """
        Get string representation of response.

        Returns:
            string representation of response.
        """
        return self.serialize()

    def __repr__(self) -> str:
        """
        Get string representation of response.

        Returns:
            string representation of response.
        """
        return str(self)



"""Client class."""
import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import requests

from manifest.request import Request

logger = logging.getLogger(__name__)


class Client(ABC):
    """Client class."""

    # Must be overridden by child class
    PARAMS: Dict[str, Tuple[str, Any]] = {}
    REQUEST_CLS = Request

    def __init__(
        self, connection_str: Optional[str] = None, client_args: Dict[str, Any] = {}
    ):
        """
        Initialize client.

        kwargs are passed to client as default parameters.

        For clients like OpenAI that do not require a connection,
        the connection_str can be None.

        Args:
            connection_str: connection string for client.
            client_args: client arguments.
        """
        self.connect(connection_str, client_args)

    @abstractmethod
    def connect(
        self, connection_str: Optional[str], client_args: Dict[str, Any]
    ) -> None:
        """
        Connect to client.

        Args:
            connection_str: connection string.
        """
        raise NotImplementedError()

    @abstractmethod
    def close(self) -> None:
        """Close the client."""
        raise NotImplementedError()

    @abstractmethod
    def get_generation_url(self) -> str:
        """Get generation URL."""
        raise NotImplementedError()

    @abstractmethod
    def get_generation_header(self) -> Dict[str, str]:
        """
        Get generation header.

        Returns:
            header.
        """
        raise NotImplementedError()

    @abstractmethod
    def supports_batch_inference(self) -> bool:
        """Return whether the client supports batch inference."""
        raise NotImplementedError()

    @abstractmethod
    def get_model_params(self) -> Dict:
        """
        Get model params.

        By getting model params from the server, we can add to request
        and make sure cache keys are unique to model.

        Returns:
            model params.
        """
        raise NotImplementedError()

    def get_model_inputs(self) -> List:
        """
        Get allowable model inputs.

        Returns:
            model inputs.
        """
        return list(self.PARAMS.keys())

    def get_request_params(
        self, prompt: Union[str, List[str]], request_args: Dict[str, Any]
    ) -> Request:
        """
        Parse model kwargs to request.

        Args:
            prompt: prompt.
            request_args: request arguments.

        Returns:
            request.
        """
        params = {"prompt": prompt}
        for key in self.PARAMS:
            params[key] = request_args.pop(key, getattr(self, key))
        return self.REQUEST_CLS(**params)

    def format_response(self, response: Dict) -> Dict[str, Any]:
        """
        Format response to dict.

        Args:
            response: response

        Return:
            response as dict
        """
        if "choices" not in response:
            raise ValueError(f"Invalid response: {response}")
        return response

    def get_request(self, request: Request) -> Tuple[Callable[[], Dict], Dict]:
        """
        Get request string function.

        Args:
            request: request.

        Returns:
            request function that takes no input.
            request parameters as dict.
        """
        if isinstance(request.prompt, list) and not self.supports_batch_inference():
            raise ValueError(
                f"{self.__class__.__name__} does not support batch inference."
            )

        request_params = request.to_dict(self.PARAMS)
        retry_timeout = request_params.pop("client_timeout")

        def _run_completion() -> Dict:
            post_str = self.get_generation_url()
            try:
                res = requests.post(
                    post_str,
                    headers=self.get_generation_header(),
                    json=request_params,
                    timeout=retry_timeout,
                )
                res.raise_for_status()
            except requests.Timeout as e:
                logger.error(
                    f"{self.__class__.__name__} request timed out."
                    " Increase client_timeout."
                )
                raise e
            except requests.exceptions.HTTPError:
                logger.error(res.json())
                raise requests.exceptions.HTTPError(res.json())
            return self.format_response(res.json())

        return _run_completion, request_params

    def get_score_prompt_request(
        self,
        request: Request,
    ) -> Tuple[Callable[[], Dict], Dict]:
        """
        Get the logit score of the prompt via a forward pass of the model.

        Args:
            request: request.

        Returns:
            request function that takes no input.
            request parameters as dict.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support prompt scoring request."
        )



"""Cache for queries and responses."""
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Union

from manifest.caches.serializers import ArraySerializer, Serializer
from manifest.response import Response

RESPONSE_CONSTRUCTORS = {
    "diffuser": {
        "generation_key": "choices",
        "logits_key": "logprobs",
        "item_key": "array",
    },
    "tomadiffuser": {
        "generation_key": "choices",
        "logits_key": "logprobs",
        "item_key": "array",
    },
}

CACHE_CONSTRUCTOR = {
    "diffuser": ArraySerializer,
    "tomadiffuser": ArraySerializer,
}


class Cache(ABC):
    """A cache for request/response pairs."""

    def __init__(
        self,
        connection_str: str,
        client_name: str = "None",
        cache_args: Dict[str, Any] = {},
    ):
        """
        Initialize client.

        Args:
            connection_str: connection string.
            client_name: name of client.
            cache_args: arguments for cache.

        cache_args are passed to client as default parameters.

        For clients like OpenAI that do not require a connection,
        the connection_str can be None.

        Args:
            connection_str: connection string for client.
            cache_args: cache arguments.
        """
        self.client_name = client_name
        self.connect(connection_str, cache_args)
        self.serializer = CACHE_CONSTRUCTOR.get(client_name, Serializer)()

    @abstractmethod
    def close(self) -> None:
        """Close the client."""
        raise NotImplementedError()

    @abstractmethod
    def connect(self, connection_str: str, cache_args: Dict[str, Any]) -> None:
        """
        Connect to client.

        Args:
            connection_str: connection string.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_key(self, key: str, table: str = "default") -> Union[str, None]:
        """
        Get the key for a request.

        With return None if key is not in cache.

        Args:
            key: key for cache.
            table: table to get key in.
        """
        raise NotImplementedError()

    @abstractmethod
    def set_key(self, key: str, value: str, table: str = "default") -> None:
        """
        Set the value for the key.

        Will override old value.

        Args:
            key: key for cache.
            value: new value for key.
            table: table to set key in.
        """
        raise NotImplementedError()

    @abstractmethod
    def commit(self) -> None:
        """Commit any results."""
        raise NotImplementedError()

    def get(
        self, request: Dict, overwrite_cache: bool, compute: Callable[[], Dict]
    ) -> Response:
        """Get the result of request (by calling compute as needed)."""
        key = self.serializer.request_to_key(request)
        cached_response = self.get_key(key)
        if cached_response and not overwrite_cache:
            cached = True
            response = self.serializer.key_to_response(cached_response)
        else:
            # Type Response
            response = compute()
            self.set_key(key, self.serializer.response_to_key(response))
            cached = False
        return Response(
            response, cached, request, **RESPONSE_CONSTRUCTORS.get(self.client_name, {})
        )
